// -*-C++-*-
//
// File:      CommentBuf.cc
// Author:    Jorgo Bakker (jbakker@astro.estec.esa.nl)
// Generated: Tue Mar  5 10:39:30 MET 2002

#include "CommentBuf.h"
#include <iostream> // show() needs std::cout!
#include <cstdlib>

CommentBuf::CommentBuf(std::streambuf* buf) :
  _realbuf(buf)
  ,_buffer(0),_bufferSize(0),_bufferFilled(0)
  ,_lineBegin(0),_lineLength(0),_lineNumber(0)
{
}

CommentBuf::~CommentBuf()
{
}

string CommentBuf::line() const
{
  if (!_buffer || !_lineLength) return string();

  // length of line must be 1 (minimum)!
  int length=_lineLength;
  if (_lineBegin[length-1] == '\n') length-=1;
  if (!length) return string();
  return string(_lineBegin, 0,length);
}


unsigned int CommentBuf::lineNumber() const
{
    return _lineNumber;
}


int CommentBuf::sync()
{
  return 0;
}


CommentBuf::int_type CommentBuf::pbackfail(int_type ch)
{
  int_type ret = traits_type::eof();
  bool testpb=eback() < gptr();
  bool testeof=ch == ret; // do we want to put back an EOF
  if (testpb) {
    if (testeof && sgetc() == ret) return ret;
    if (seekoff(-1,std::ios_base::cur) != pos_type(-1)) return ch;
  }
  return ret;

}


CommentBuf::int_type CommentBuf::underflow()
{
  _lineBegin = egptr(); // point the next line to the end of current buffer.
  char_type* ptr = _lineBegin;
  
  /* First check whether we already have read stuff into the buffer!
     (This may happen if we have rewinded the buffer) */
  while (ptr < eback() + _bufferFilled) {
    if (*ptr++ == '\n') {
      setg(eback(),_lineBegin,ptr);
      _lineLength = ptr - _lineBegin;
      _lineNumber++;
      return _lineBegin[0];
    }
  }

  /* 
     eptr points to end of line!
  */
  
  
  /* Copy a line from streambuf to buffer */
  int_type ch;
  for (;;) {
    ch = _realbuf->sbumpc();          // Get next char from streambuf
    
    if (ch == EOF) return EOF;
      
    // skip comment on this line!
    if (ch == '#') {             // Comment?
      while (((ch = _realbuf->sbumpc()) != '\n') && (ch != EOF));
      if (ch == EOF) return EOF;
    }
    
    // store
    if (ptr == egptr()) ptr = enlargeBuffer();
    *ptr++ = ch;
    _bufferFilled++;

    if (ch == '\n') break;
  }
  
  /* now we have read a complete line, set [gptr(),eptr()]=[BOL,EOL]*/
  setg(eback(),_lineBegin,ptr);
  _lineLength = ptr - _lineBegin;
  _lineNumber++;
  
  return (_lineBegin?_lineBegin[0]:EOF);
}

/*
 *  Enlarge the buffer and return pointer to previous end point.
 *  Hence the buffer dynamically resizes to accomodate the longest line.
 */

CommentBuf::char_type* CommentBuf::enlargeBuffer()
{
  if (egptr() < _buffer + _bufferSize) {
    setg(_buffer,egptr(),_buffer + _bufferSize);
  }
  else {
    // create a new buffer
    unsigned int newSize=0;
    if (_bufferSize) newSize=_bufferSize*2;
    else newSize = 1024;
    char_type* newBuffer = new char_type[newSize];

    // copy old stuff
    if (_bufferSize) memcpy(newBuffer, _buffer, _bufferSize);

    // replace and delete _buffer
    setbuf(newBuffer,newSize);
    setg(newBuffer,newBuffer+_bufferSize,newBuffer+newSize);
    _lineBegin=newBuffer+(_lineBegin-_buffer);
    
    delete [] _buffer;
    _buffer=newBuffer;
    _bufferSize=newSize;
  }

  return gptr();
}


/* 
   Assumptions:
   * gptr, egptr, _lineNumber, and _lineBegin are consistent before entry
   * pos in [0,_bufferFilled], where _bufferFilled >0!   
*/
void CommentBuf::findLineInfo(CommentBuf::pos_type pos)
{
  char_type* address=_buffer+pos;

  // return immeditially in this Line!
  if (address >= _lineBegin && address <= egptr()) {
    setg(_buffer,_buffer+pos,egptr());
    return;
  }

  char_type* ptr=egptr();
  
  if (address>ptr) {
    // forward:
    while (address>ptr) {
      _lineNumber++;
      _lineBegin=ptr++;

      // set ptr to one past EOL
      while (*ptr != '\n' && ptr < _buffer + _bufferFilled ) ptr++;
    }
  }
  else {
    // rewind:
    if (_lineBegin ==  _buffer + _bufferFilled)
      _lineNumber++; // fake a line at end of file!

    while (address < _lineBegin) {
      ptr=_lineBegin;
      if (*(--_lineBegin) == '\n') _lineNumber--;
      while (_lineBegin > _buffer && *(_lineBegin-1) != '\n') _lineBegin--;
    }
  }

  // set it:
  setg(_buffer,address,ptr);
  _lineLength=ptr-_lineBegin;
}


/*
  Flaw here!

  return=is.streambuf->seekpos(pos);
  intern=is.tellg();
  state=!is;


  pos<0:
          GCC-2.95  GCC-3.0  ICC-5.0 CommentBuf
  return  LAST      LAST     -1      -1
  state   BAD       GOOD     BAD     BAD


  pos>EOF:
          GCC-2.95  GCC-3.0  ICC-5.0 CommentBuf
  return  pos       pos      pos     -1
  state   GOOD      GOOD     GOOD    BAD

 */
CommentBuf::pos_type
CommentBuf::seekpos(CommentBuf::pos_type pos,std::ios_base::openmode)
{
  // don't bother if:
  pos_type ret=pos_type(-1);
  if (pos < 0)  {
    return ret;
  }

  // setup buffer if necessary
  if (!_buffer) underflow();

  if (pos == pos_type(0)) return pos;

  //char_type* address=_buffer+pos;
  if (!_buffer) underflow();

  // populate the buffer, if necessary:
  while ((unsigned int)pos > _bufferFilled) {
    findLineInfo(_bufferFilled);
    if (snextc() == EOF) {
      return ret;
    }
  }
    
  findLineInfo(pos);
  return pos;
}

CommentBuf::pos_type
CommentBuf::seekoff(off_type off,
		    std::ios_base::seekdir way,
		    std::ios_base::openmode m) 
{
  pos_type pos;
  if (way == std::ios_base::end) {
    while (snextc() != EOF) {
      findLineInfo(_bufferFilled);
    }
    pos=_bufferFilled+off;
  }
  else if (way == std::ios_base::beg) pos=off;
  else pos=(gptr()-eback())+off;
  return seekpos(pos,m);

}


/* For debug purposes only */
void
CommentBuf::show()
{
  char_type c=traits_type::to_char_type(sgetc());
  std::cout
    << "sgetc(";
  if (c == '\n')  	  std::cout << "\\n";
  else if (sgetc() == EOF)  std::cout << "EOF";
  else std::cout << c;
  std::cout
    << "), "
    /*
      << "eback(" << int(eback()) << "),"
      << "gptr("  << int(gptr()) << "),"
      << "egptr(" << int(egptr())<< "),"
    */
    << "rel: "
    << "gptr("  << int(gptr()-eback())<< "), "
    << "egptr(" << int(egptr()-eback())<< "), "
    << "lbegin(" << int(_lineBegin-eback()) << "), "
    << "line(" << _lineNumber<< "), "
    << "buffer(" << _bufferSize << "), "
    << std::endl;
}
